@@ -7,6 +7,7 @@ from collections import defaultdict
 
 import certifi
 import tenacity
+import requests
 from json_home_client import Client as APIClient
 
 from .. import config, messages as m, t
@@ -28,8 +29,8 @@ if t.TYPE_CHECKING:
         text: str
         spec: str
 
-    RawAnchorT = t.TypedDict(
-        "RawAnchorT",
+    ShepherdAnchorT = t.TypedDict(
+        "ShepherdAnchorT",
         {
             "name": t.Required[str],
             "type": t.Required[str],
@@ -45,6 +46,27 @@ if t.TYPE_CHECKING:
         total=False,
     )
 
+    WebrefAnchorT = t.TypedDict(
+        "WebrefAnchorT",
+        {
+            "id": t.Required[str],
+            "type": t.Required[str],
+            "for": list[str],
+            "informative": bool,
+            "access": t.Required[str],
+            "linkingText": list[str],
+            "href": t.Required[str],
+        },
+        total=False,
+    )
+
+    class WebrefHeadingT(t.TypedDict, total=False):
+        id: t.Required[str]
+        href: t.Required[str]
+        title: t.Required[str]
+        level: t.Required[int]
+        number: str
+
     # Need to use function form due to "for" key
     # being invalid as a property name
     AnchorT = t.TypedDict(
@@ -62,7 +84,7 @@ if t.TYPE_CHECKING:
         },
     )
 
-    class RawSpecT(t.TypedDict, total=False):
+    class ShepherdSpecT(t.TypedDict, total=False):
         name: t.Required[str]
         short_name: t.Required[str]
         title: t.Required[str]
@@ -72,6 +94,22 @@ if t.TYPE_CHECKING:
         draft_uri: str | None
         abstract: str | None
 
+    class WebrefSpecT(t.TypedDict, total=False):
+        url: t.Required[str]
+        shortname: t.Required[str]
+        title: t.Required[str]
+        shortTitle: t.Required[str]
+        series: t.Required[dict[str, str]]
+        nightly: t.Required[dict[str, str | list[str]]]
+        release: dict[str, str | list[str]]
+        seriesVersion: str
+        links: str
+        refs: str
+        idl: str
+        dfns: str
+        headings: str
+        ids: str
+
     class SpecT(t.TypedDict):
         vshortname: str
         shortname: str
@@ -88,57 +126,13 @@ def progressMessager(index: int, total: int) -> t.Callable[[], None]:
 
 
 def update(path: str, dryRun: bool = False) -> set[str] | None:
-    m.say("Downloading anchor data...")
-    shepherd = APIClient(
-        "https://api.csswg.org/shepherd/",
-        version="vnd.csswg.shepherd.v1",
-        ca_cert_path=certifi.where(),
-    )
-    rawSpecData = dataFromApi(shepherd, "specifications", draft=True)
-    if not rawSpecData:
-        return None
-
     specs: SpecsT = dict()
     anchors: AnchorsT = defaultdict(list)
     headings: AllHeadingsT = {}
-    lastMsgTime: float = 0
-    for i, rawSpec in enumerate(rawSpecData.values(), 1):
-        lastMsgTime = config.doEvery(
-            s=5,
-            lastTime=lastMsgTime,
-            action=progressMessager(i, len(rawSpecData)),
-        )
-        rawSpec = dataFromApi(shepherd, "specifications", draft=True, anchors=True, spec=rawSpec["name"])
-        spec = genSpec(rawSpec)
-        assert spec["vshortname"] is not None
-        specs[spec["vshortname"]] = spec
-        specHeadings: HeadingsT = {}
-        headings[spec["vshortname"]] = specHeadings
-
-        def setStatus(obj: RawAnchorT, status: str) -> RawAnchorT:
-            obj["status"] = status
-            return obj
-
-        rawAnchorData = [setStatus(x, "snapshot") for x in linearizeAnchorTree(rawSpec.get("anchors", []))] + [
-            setStatus(x, "current") for x in linearizeAnchorTree(rawSpec.get("draft_anchors", []))
-        ]
-        for rawAnchor in rawAnchorData:
-            rawAnchor = fixupAnchor(rawAnchor)
-            linkingTexts = rawAnchor["linking_text"]
-            assert linkingTexts is not None
-            if len(linkingTexts) == 0:
-                # Happens if it had no linking text at all originally
-                continue
-            if len(linkingTexts) == 1 and linkingTexts[0].strip() == "":
-                # Happens if it was marked with an empty lt and Shepherd still picked it up
-                continue
-            if "section" in rawAnchor and rawAnchor["section"] is True:
-                addToHeadings(rawAnchor, specHeadings, spec=spec)
-            if rawAnchor["type"] in config.dfnTypes.union(["dfn"]):
-                addToAnchors(rawAnchor, anchors, spec=spec)
 
+    gatherWebrefData(specs, anchors, headings)
+    gatherShepherdData(specs, anchors, headings)
     cleanSpecHeadings(headings)
-
     methods = extractMethodData(anchors)
     fors = extractForsData(anchors)
 
@@ -187,6 +181,117 @@ def update(path: str, dryRun: bool = False) -> set[str] | None:
     return writtenPaths
 
 
+def gatherWebrefData(specs: SpecsT, anchors: AnchorsT, headings: AllHeadingsT) -> None:
+    m.say("Downloading anchor data from Webref...")
+    currentWebrefData = specsFromWebref("current")
+    snapshotWebrefData = specsFromWebref("snapshot")
+
+    lastMsgTime: float = 0
+    for i, rawWSpec in enumerate(currentWebrefData, 1):
+        lastMsgTime = config.doEvery(
+            s=5,
+            lastTime=lastMsgTime,
+            action=progressMessager(i, len(currentWebrefData)),
+        )
+        if "dfns" not in rawWSpec and "headings" not in rawWSpec:
+            continue
+        spec = genWebrefSpec(rawWSpec)
+        currentUrl = spec["current_url"]
+        assert currentUrl is not None
+
+        specs[spec["vshortname"].lower()] = spec
+        specHeadings: HeadingsT = {}
+        headings[spec["vshortname"]] = specHeadings
+        rawAnchorData: list[ShepherdAnchorT] = []
+
+        if "dfns" in rawWSpec:
+            currentAnchors = anchorsFromWebref("current", rawWSpec["dfns"])
+            if currentAnchors:
+                rawAnchorData = [convertWebrefAnchor(x, currentUrl, "current") for x in currentAnchors]
+        if "headings" in rawWSpec:
+            currentHeadings = headingsFromWebref("current", rawWSpec["headings"])
+            if currentHeadings:
+                rawAnchorData += [convertWebrefHeading(x, currentUrl, "current") for x in currentHeadings]
+
+        # Complete list of anchors/headings with those from the snapshot version of the spec
+        if spec["snapshot_url"] is not None and snapshotWebrefData is not None:
+            rawSnapshotSpec: WebrefSpecT | None = None
+            for s in snapshotWebrefData:
+                if s["shortname"].lower() == spec["vshortname"].lower():
+                    rawSnapshotSpec = s
+                    break
+            else:
+                print(f"Despite claiming to have a snapshot url, no snapshot data found for '{spec['vshortname']}'.")
+            if rawSnapshotSpec:
+                if "dfns" in rawSnapshotSpec:
+                    snapshotAnchors = anchorsFromWebref("snapshot", rawSnapshotSpec["dfns"])
+                    if snapshotAnchors:
+                        rawAnchorData += [
+                            convertWebrefAnchor(x, spec["snapshot_url"], "snapshot") for x in snapshotAnchors
+                        ]
+                if "headings" in rawSnapshotSpec:
+                    snapshotHeadings = headingsFromWebref("snapshot", rawSnapshotSpec["headings"])
+                    if snapshotHeadings:
+                        rawAnchorData += [
+                            convertWebrefHeading(x, spec["snapshot_url"], "snapshot") for x in snapshotHeadings
+                        ]
+
+        if len(rawAnchorData) > 0:
+            processRawAnchors(rawAnchorData, anchors, specHeadings, spec)
+
+
+def gatherShepherdData(specs: SpecsT, anchors: AnchorsT, headings: AllHeadingsT) -> None:
+    m.say("Downloading anchor data from Shepherd...")
+    shepherd = APIClient(
+        "https://api.csswg.org/shepherd/",
+        version="vnd.csswg.shepherd.v1",
+        ca_cert_path=certifi.where(),
+    )
+    rawSpecData = dataFromApi(shepherd, "specifications", draft=True)
+    if not rawSpecData:
+        return None
+
+    neededShepherdSpecs = [
+        "css-2022",
+        "css-color-3",
+        "css-color-6",
+        "css-color-hdr-1",
+        "css-conditional-values-1",
+        "css-contain-1",
+        "css-display-4",
+        "css-forms-1",
+        "css-grid-1",
+        "css-ui-3",
+        "css-writing-modes-3",
+        "dom-level-2-style",
+        "mediaqueries-3",
+    ]
+    lastMsgTime: float = 0
+    for i, rawSSpec in enumerate(rawSpecData.values(), 1):
+        lastMsgTime = config.doEvery(
+            s=5,
+            lastTime=lastMsgTime,
+            action=progressMessager(i, len(rawSpecData)),
+        )
+        if rawSSpec["name"] not in neededShepherdSpecs:
+            continue
+        if isSpecInList(genSpec(rawSSpec), specs):
+            # Gradually stop as they get filled in
+            continue
+        print(rawSSpec["name"])
+        rawSSpec = dataFromApi(shepherd, "specifications", draft=False, anchors=False, spec=rawSSpec["name"])
+        spec = genSpec(rawSSpec)
+        assert spec["vshortname"] is not None
+        specs[spec["vshortname"]] = spec
+        specHeadings: HeadingsT = {}
+        headings[spec["vshortname"]] = specHeadings
+
+        rawAnchorData = [setStatus(x, "snapshot") for x in linearizeAnchorTree(rawSSpec.get("anchors", []))] + [
+            setStatus(x, "current") for x in linearizeAnchorTree(rawSSpec.get("draft_anchors", []))
+        ]
+        processRawAnchors(rawAnchorData, anchors, specHeadings, spec)
+
+
 @tenacity.retry(reraise=True, stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(1, 2))
 def dataFromApi(api: APIClient, *args: t.Any, **kwargs: t.Any) -> t.JSONT:
     anchorDataContentTypes = [
@@ -214,7 +319,47 @@ def dataFromApi(api: APIClient, *args: t.Any, **kwargs: t.Any) -> t.JSONT:
     return t.cast("t.JSONT", data)
 
 
-def linearizeAnchorTree(multiTree: list, rawAnchors: list[dict[str, t.Any]] | None = None) -> list[RawAnchorT]:
+def specsFromWebref(status: t.Literal["current" | "snapshot"]) -> list[WebrefSpecT]:
+    url = ("ed" if status == "current" else "tr") + "/index.json"
+    j = dataFromWebref(url)
+    if j is None or j.get("results") is None:
+        raise Exception(f"No {status} specs data from WebRef. Got:\n{json.dumps(j, indent=1)}")
+    return t.cast("list[WebrefSpecT]", j["results"])
+
+
+def anchorsFromWebref(status: t.Literal["current" | "snapshot"], urlSuffix: str) -> list[WebrefAnchorT]:
+    url = ("ed" if status == "current" else "tr") + "/" + urlSuffix
+    j = dataFromWebref(url)
+    if j is None or j.get("dfns") is None:
+        raise Exception(f"No WebRef dfns data at {url}. Got:\n{json.dumps(j, indent=1)}")
+    return t.cast("list[WebrefAnchorT]", j["dfns"])
+
+
+def headingsFromWebref(status: t.Literal["current" | "snapshot"], urlSuffix: str) -> list[WebrefHeadingT]:
+    url = ("ed" if status == "current" else "tr") + "/" + urlSuffix
+    j = dataFromWebref(url)
+    if j is None or j.get("headings") is None:
+        raise Exception(f"No WebRef headings data at {url}. Got:\n{json.dumps(j, indent=1)}")
+    return t.cast("list[WebrefHeadingT]", j["headings"])
+
+
+@tenacity.retry(reraise=True, stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(1, 2))
+def dataFromWebref(url: str) -> t.JSONT:
+    webrefAPIUrl = "https://raw.githubusercontent.com/w3c/webref/main/"
+    try:
+        response = requests.get(webrefAPIUrl + url)
+    except Exception as e:
+        raise Exception(f"Couldn't download data from Webref.\n{e}") from e
+    try:
+        data = response.json()
+    except Exception as e:
+        raise Exception(
+            f"Data retrieved from Webref wasn't valid JSON for some reason. Try downloading again?\n{e}"
+        ) from e
+    return t.cast("t.JSONT", data)
+
+
+def linearizeAnchorTree(multiTree: list, rawAnchors: list[dict[str, t.Any]] | None = None) -> list[ShepherdAnchorT]:
     if rawAnchors is None:
         rawAnchors = []
     # Call with multiTree being a list of trees
@@ -224,10 +369,10 @@ def linearizeAnchorTree(multiTree: list, rawAnchors: list[dict[str, t.Any]] | No
         if item.get("children"):
             linearizeAnchorTree(item["children"], rawAnchors)
             del item["children"]
-    return t.cast("list[RawAnchorT]", rawAnchors)
+    return t.cast("list[ShepherdAnchorT]", rawAnchors)
 
 
-def genSpec(rawSpec: RawSpecT) -> SpecT:
+def genSpec(rawSpec: ShepherdSpecT) -> SpecT:
     assert rawSpec["name"] is not None
     assert rawSpec["short_name"] is not None
     assert rawSpec["title"] is not None
@@ -264,7 +409,71 @@ def genSpec(rawSpec: RawSpecT) -> SpecT:
     return spec
 
 
-def fixupAnchor(anchor: RawAnchorT) -> RawAnchorT:
+def genWebrefSpec(rawSpec: WebrefSpecT) -> SpecT:
+    """Generate a spec object from data gleaned from Webref"""
+    assert rawSpec["shortname"] is not None
+    assert rawSpec["series"] is not None
+    assert rawSpec["nightly"] is not None
+    assert rawSpec["title"] is not None
+    assert rawSpec["shortTitle"] is not None
+    spec: SpecT = {
+        "vshortname": rawSpec["shortname"].lower(),
+        "shortname": rawSpec["series"]["shortname"].lower(),
+        "snapshot_url": t.cast("str", rawSpec.get("release", {}).get("url")),
+        "current_url": t.cast("str", rawSpec["nightly"]["url"]),
+        "title": rawSpec["shortTitle"],
+        "description": rawSpec["title"],
+        "abstract": None,
+        "level": int(rawSpec["seriesVersion"])
+        if "seriesVersion" in rawSpec and re.match(r"^\d+$", rawSpec.get("seriesVersion", ""))
+        else None,
+    }
+    return spec
+
+
+def convertWebrefHeading(heading: WebrefHeadingT, specUrl: str, status: str) -> ShepherdAnchorT:
+    """Convert a heading returned by Webref to the anchor format used in Shepherd"""
+    assert heading["id"] is not None
+    assert heading["title"] is not None
+    assert heading["href"] is not None
+    anchor: ShepherdAnchorT = {
+        "name": heading["number"] if "number" in heading else heading["id"],
+        "type": "heading",
+        "for": [],
+        "section": True,
+        "title": heading["title"],
+        "status": status,
+        "normative": True,
+        "export": True,
+        "linking_text": [heading["title"]],
+        "uri": heading["href"].replace(specUrl, ""),
+    }
+    return anchor
+
+
+def convertWebrefAnchor(rawAnchor: WebrefAnchorT, specUrl: str, status: str) -> ShepherdAnchorT:
+    """Convert an anchor returned by Webref to the anchor format used in Shepherd"""
+    assert rawAnchor["id"] is not None
+    assert rawAnchor["type"] is not None
+    assert rawAnchor["linkingText"] is not None
+    assert rawAnchor["access"] is not None
+    assert rawAnchor["href"] is not None
+    anchor: ShepherdAnchorT = {
+        "name": rawAnchor["id"],
+        "type": rawAnchor["type"],
+        "for": rawAnchor["for"],
+        "section": False,
+        "title": rawAnchor["linkingText"][0],
+        "status": status,
+        "normative": not rawAnchor["informative"],
+        "export": rawAnchor["access"] == "public",
+        "linking_text": rawAnchor["linkingText"],
+        "uri": rawAnchor["href"].replace(specUrl, ""),
+    }
+    return anchor
+
+
+def fixupAnchor(anchor: ShepherdAnchorT) -> ShepherdAnchorT:
     """Miscellaneous fixes to the anchors before I start processing"""
 
     # This one issue was annoying
@@ -302,7 +511,7 @@ def fixupAnchor(anchor: RawAnchorT) -> RawAnchorT:
     return anchor
 
 
-def addToHeadings(rawAnchor: RawAnchorT, specHeadings: HeadingsT, spec: SpecT) -> None:
+def addToHeadings(rawAnchor: ShepherdAnchorT, specHeadings: HeadingsT, spec: SpecT) -> None:
     uri = rawAnchor["uri"]
     assert uri is not None
     if rawAnchor["status"] == "snapshot":
@@ -367,7 +576,7 @@ def cleanSpecHeadings(headings: AllHeadingsT) -> None:
                 del specHeadings[v[0]]
 
 
-def addToAnchors(rawAnchor: RawAnchorT, anchors: AnchorsT, spec: SpecT) -> None:
+def addToAnchors(rawAnchor: ShepherdAnchorT, anchors: AnchorsT, spec: SpecT) -> None:
     if rawAnchor["status"] == "snapshot":
         baseUrl = spec["snapshot_url"]
     else:
@@ -391,6 +600,25 @@ def addToAnchors(rawAnchor: RawAnchorT, anchors: AnchorsT, spec: SpecT) -> None:
         anchors[text].append(anchor)
 
 
+def processRawAnchors(
+    rawAnchorData: list[ShepherdAnchorT], anchors: AnchorsT, specHeadings: HeadingsT, spec: SpecT
+) -> None:
+    for rawAnchor in rawAnchorData:
+        rawAnchor = fixupAnchor(rawAnchor)
+        linkingTexts = rawAnchor["linking_text"]
+        assert linkingTexts is not None
+        if len(linkingTexts) == 0:
+            # Happens if it had no linking text at all originally
+            continue
+        if len(linkingTexts) == 1 and linkingTexts[0].strip() == "":
+            # Happens if it was marked with an empty lt and Shepherd still picked it up
+            continue
+        if "section" in rawAnchor and rawAnchor["section"] is True:
+            addToHeadings(rawAnchor, specHeadings, spec=spec)
+        if rawAnchor["type"] in config.dfnTypes.union(["dfn"]):
+            addToAnchors(rawAnchor, anchors, spec=spec)
+
+
 def extractMethodData(anchors: AnchorsT) -> MethodsT:
     """Compile a db of {argless methods => {argfull method => {args, fors, url, shortname}}"""
 
@@ -437,6 +665,23 @@ def extractForsData(anchors: AnchorsT) -> ForsT:
     return fors
 
 
+def setStatus(obj: ShepherdAnchorT, status: str) -> ShepherdAnchorT:
+    obj["status"] = status
+    return obj
+
+
+def isSpecInList(spec: SpecT, specs: SpecsT) -> bool:
+    if spec["vshortname"].lower() in specs:
+        return True
+    for s in specs.values():
+        # Shortnames don't always match. Let's compare URLs
+        if (s["snapshot_url"] is not None and s["snapshot_url"] == spec["snapshot_url"]) or (
+            s["current_url"] is not None and s["current_url"] == spec["current_url"]
+        ):
+            return True
+    return False
+
+
 def writeAnchorsFile(anchors: AnchorsT, path: str) -> set[str]:
     """
     Keys may be duplicated.
@@ -463,7 +708,7 @@ def writeAnchorsFile(anchors: AnchorsT, path: str) -> set[str]:
         writtenPaths.add(p)
         with open(p, "w", encoding="utf-8") as fh:
             for key, entries in sorted(group_anchors.items(), key=lambda x: x[0]):
-                for e in entries:
+                for e in sorted(entries, key=lambda x: x["url"]):
                     fh.write(key + "\n")
                     for field in [
                         "type",
